unit MineSweeper;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, LResources, Forms, Controls, Graphics, Dialogs, ExtCtrls,
  Buttons, StdCtrls, Spin;
const
  FieldSizeH = 14;
  FieldSizeHPlus1 = FieldSizeH + 1;
  FieldSizeW1 = 16;
  FieldSizeW2 = 24;
  FieldSizeW2Plus1 = FieldSizeW2 + 1;
  CellDispSize = 16;
  MainPaintBoxHeight = FieldSizeH * CellDispSize;
  mnum1 = 30;
  mnum2 = 60;
  clOrange = TColor($0080FF);
type
  CellStatus = (isMine, isMarked, isRevealed);
  CellStatusSet = set of CellStatus;
  MineCell = record
    status    : CellStatusSet;
    neighbour : byte
  end;
  MineField = array[0..FieldSizeW2Plus1, 0..FieldSizeHPlus1] of MineCell;
  
  { TForm1 }

  TForm1 = class(TForm)
    WideCheckBox: TCheckBox;
    DoneButton: TButton;
    QuitButton: TButton;
    ResetButton: TButton;
    MineNumSpinEdit: TSpinEdit;
    Label1: TLabel;
    Label2: TLabel;
    ResponseLabel: TLabel;
    TimeLabel: TLabel;
    RemainLabel: TLabel;
    MainPaintBox: TPaintBox;
    MarkToggleBox: TToggleBox;
    Timer1: TTimer;
    procedure DoneButtonClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure MainPaintBoxMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure MainPaintBoxPaint(Sender: TObject);
    procedure QuitButtonClick(Sender: TObject);
    procedure ResetButtonClick(Sender: TObject);
    procedure Timer1Timer(Sender: TObject);
    procedure WideCheckBoxChange(Sender: TObject);
  private
    { private declarations }
    mfld : MineField;
    mnum : integer;
    remain : integer;
    timepast : word;
    running : boolean;
    FieldSizeW,
    MainPaintBoxWidth : integer;
    procedure InitVars;
    procedure InitMF;
    procedure RevealMF(var g : TCanvas);
    procedure NormalMFView(var g : TCanvas);
    procedure revealblock(x, y : integer);
  public
    { public declarations }
  end; 

var
  Form1: TForm1; 

implementation

uses AboutForm;

function fnstr(i : integer) : string;
var s : string;
begin
  str(i, s);
  fnstr := s
end;

{ TForm1 }
procedure TForm1.InitVars;
begin
  if WideCheckBox.State = cbUnchecked then FieldSizeW := FieldSizeW1
                                      else FieldSizeW := FieldSizeW2;
  MainPaintBoxWidth := FieldSizeW * CellDispSize;
  MainPaintBox.Width := succ(MainPaintBoxWidth);
  Width := MainPaintBoxWidth + 16;
  InitMF;
  RemainLabel.caption := fnstr(remain);
  ResponseLabel.Caption := 'GoOn!';
  timepast := 0;
  running := true;
  Timer1.Enabled := true
end;

procedure TForm1.FormCreate(Sender: TObject);
var x, y : integer;
begin
  randomize;
  fillchar(mfld, sizeof(mfld), 0);
  FieldSizeW := FieldSizeW1;
  MainPaintBoxWidth := FieldSizeW * CellDispSize;
  for x := 1 to FieldSizeW do
    for y := 1 to FieldSizeH do
      include(mfld[x, y].status, isRevealed);
  running := true
end;

procedure TForm1.DoneButtonClick(Sender: TObject);
var n, x, y : integer;
    s       : string;
begin
  Timer1.Enabled := false;
  running := false;
  n := mnum;
  for x := 1 to FieldSizeW do
    for y := 1 to FieldSizeH do
      dec(n, ord(mfld[x, y].status * [isMine, isMarked] = [isMine, isMarked]));
  if n = 0 then s := 'Great!' else s := 'Failed...';
  ResponseLabel.Caption := s;
  MainPaintBox.refresh
end;

procedure TForm1.MainPaintBoxMouseDown(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
  if not Timer1.Enabled then Form2.Show;
  x := succ(x div CellDispSize);
  y := succ(y div CellDispSize);
  if (x in [1..FieldSizeW]) and (y in [1..FieldSizeH]) then begin
    with mfld[x, y] do
      if (Button = mbRight) or MarkToggleBox.checked then begin
        if isMarked in status then begin
          exclude(status, isMarked);
          inc(remain)
        end else begin
          include(status, isMarked);
          dec(remain)
        end;
        DoneButton.Enabled := (remain = 0)
      end else
        if isMine in status then begin
          running := false;
          Timer1.Enabled := false;
          remain := mnum;
          ResponseLabel.Caption := 'Oops...'
        end else
          revealblock(x, y);
    MainPaintBox.refresh
  end;
  RemainLabel.caption := fnstr(remain)
end;

procedure TForm1.revealblock(x, y : integer);
begin
  with mfld[x, y] do
    if status * [isMine, isRevealed] = [] then begin
      include(status, isRevealed);
      if neighbour = 0 then begin
        revealblock(x-1, y-1);
        revealblock(x-1, y  );
        revealblock(x-1, y+1);
        revealblock(x  , y-1);
        revealblock(x  , y+1);
        revealblock(x+1, y-1);
        revealblock(x+1, y  );
        revealblock(x+1, y+1)
      end
    end
end;

procedure TForm1.NormalMFView(var g : TCanvas);
var x, y, xx, yy : integer;
begin
  with g do begin
    Pen.Color := clBlack;
    xx := 0;
    for x := 1 to FieldSizeW do begin
      yy := 0;
      for y := 1 to FieldSizeH do begin
        with mfld[x, y] do
          if isMarked in status then begin
            Brush.Color := clYellow;
            FillRect(xx, yy, xx + CellDispSize, yy + CellDispSize)
          end else if isRevealed in status then begin
            Brush.Color := clWhite;
            FillRect(xx, yy, xx + CellDispSize, yy + CellDispSize);
            if neighbour > 0 then TextOut(succ(xx), succ(yy), fnstr(neighbour))
          end;
        inc(yy, CellDispSize)
      end;
      inc(xx, CellDispSize)
    end
  end
end;

procedure TForm1.RevealMF(var g : TCanvas);
var x, y, xx, yy : integer;
begin
  with g do begin
    Brush.Color := clWhite;
    FillRect(0, 0, MainPaintBoxWidth, MainPaintBoxHeight);
    Pen.Color := clBlack;
    Brush.Color := clRed;
    xx := 0;
    for x := 1 to FieldSizeW do begin
      xx := pred(x) * CellDispSize;
      yy := 0;
      for y := 1 to FieldSizeH do begin
        with mfld[x, y] do
          if isMine in status then begin
            if isMarked in status then Brush.Color := clOrange
                                  else Brush.Color := clRed;
            FillRect(xx, yy, xx + CellDispSize, yy + CellDispSize)
          end else begin
            if isMarked in status then begin
              Brush.Color := clYellow;
              FillRect(xx, yy, xx + CellDispSize, yy + CellDispSize)
            end;
            if neighbour > 0 then TextOut(succ(xx), succ(yy), fnstr(neighbour))
          end;
        inc(yy, CellDispSize)
      end;
      inc(xx, CellDispSize)
    end
  end
end;

procedure TForm1.MainPaintBoxPaint(Sender: TObject);
var x, y, xx, yy : integer;
begin
  if running then NormalMFView(MainPaintBox.Canvas)
             else RevealMF(MainPaintBox.Canvas);
  with MainPaintBox.Canvas do begin
    Pen.Color := clBlack;
    for x := 0 to FieldSizeW do begin
      xx := x * CellDispSize;
      line(xx, 0, xx,  MainPaintBoxHeight)
    end;
    for y := 0 to FieldSizeH do begin
      yy := y * CellDispSize;
      line(0, yy, MainPaintBoxWidth, yy)
    end
  end
end;

procedure TForm1.QuitButtonClick(Sender: TObject);
begin
  close
end;

procedure TForm1.ResetButtonClick(Sender: TObject);
begin
  InitVars;
  MainPaintBox.refresh
end;

procedure TForm1.Timer1Timer(Sender: TObject);
begin
  inc(timepast);
  TimeLabel.Caption := fnstr(timepast)
end;

procedure TForm1.WideCheckBoxChange(Sender: TObject);
begin
  if WideCheckBox.State = cbUnchecked then MineNumSpinEdit.Value := mnum1
                                      else MineNumSpinEdit.Value := mnum2
end;

procedure TForm1.InitMF;
var i, x, y : integer;
begin
  fillchar(mfld, sizeof(mfld), 0);
  mnum := MineNumSpinEdit.Value;
  remain := mnum;
  for i := 1 to mnum do begin
    repeat
      x := succ(random(FieldSizeW));
      y := succ(random(FieldSizeH))
    until not (isMine in mfld[x, y].status);
    include(mfld[x, y].status, isMine);
    inc(mfld[x-1, y-1].neighbour);
    inc(mfld[x-1, y  ].neighbour);
    inc(mfld[x-1, y+1].neighbour);
    inc(mfld[x  , y-1].neighbour);
    inc(mfld[x  , y+1].neighbour);
    inc(mfld[x+1, y-1].neighbour);
    inc(mfld[x+1, y  ].neighbour);
    inc(mfld[x+1, y+1].neighbour)
  end;
  for i := 0 to succ(FieldSizeW) do begin     (* sentinels *)
    mfld[i, 0].neighbour := 255;
    mfld[i, FieldSizeHPlus1].neighbour := 255
  end;
  for i := 0 to FieldSizeHPlus1 do begin     (* sentinels *)
    mfld[0, i].neighbour := 255;
    mfld[succ(FieldSizeW), i].neighbour := 255
  end
end;

initialization
  {$I minesweeper.lrs}

end.

